# -*- coding: utf-8 -*-
# Project : flacon
# File : converter.py
# (c) Alexander Sokolov asokol@mail.ru
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


#import errno
#import exceptions
import re
import sys
import time


from PyQt4 import QtCore
from PyQt4.QtCore import *

from flaconglobal import *
from disc import *
from errors import *


class Converter(QObject):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self, parent = None):
        QObject.__init__(self, parent)
        self._disc = None;
        self._threads = []
        self._threadCount = 0
        self._startedTime = 0


   #****************************************************************
   #
   #****************************************************************
    def start(self, disc):
        self._startedTime = time.time()
        self._disc = disc
        format = disc.outFormat()

        try:
            encoderClass = getEncoderClass(format)
            albumGainClass = getAlbumGainClass(format)
            trackGainClass = getTrackGainClass(format)
            self.check([Splitter, encoderClass, albumGainClass, trackGainClass])

        except Exception,e:
            self.emit(SIGNAL("error(QString)"), QString(unicode(e)))
            self.emit(SIGNAL("finished()"))
            return


        (self._threadCount, ok) = qApp.preferences.value('Encoder/ThreadCount').toInt()

        if self._threadCount<1:
            self._threadCount = max(6, QThread.idealThreadCount())


        # Add splitter ...............................
        splitter = Splitter(self._disc)
        self._threads.append(splitter)

        #Add AlbumGain ...............................
        albumGain = None
        if albumGainClass:
            albumGain = albumGainClass(range(0, self._disc.trackCount()))
            self._threads.append(albumGain)

        for track in self._disc.tracks():
            encoder = encoderClass()
            encoder.setTrack(track)
            self._threads.append(encoder)
            self.connect(splitter, SIGNAL("trackReady(int, QString)"), encoder.inputDataReady)

            #  CREATE THREADS CHAINS **********************************************
            #
            #        SIGNAL(trackReady) --> SLOT(inputDataReady)
            #
            #             +--> Encoder ----> Track gain -->+
            #  Splitter ->+              ...               +-> Album gain ---> self
            #             +--> Encoder ----> Track gain -->+
            #
            #                                optional step    optional step
            #**********************************************************************
            steps = []
            steps.append(encoder)

            if trackGainClass:
                trackGain = trackGainClass([track.index()])
                self._threads.append(trackGain)
                steps.append(trackGain)

            if albumGain:
                steps.append(albumGain)

            steps.append(self)
            # ........................................


            prev = encoder
            for next in steps:
                self.connect(prev, SIGNAL("trackReady(int, QString)"), next.inputDataReady)
                prev = next


        for thread in self._threads:
            self.connect(thread, SIGNAL("error(int, QString)"), self.threadError)
            self.connect(thread, SIGNAL("readyStart()"), self.startThread)
            self.connect(thread, SIGNAL("finished()"), self.threadFinished)
            self.connect(thread, SIGNAL("progress(int, int, int)"), self._disc.progressChanged)

        splitter.start()


   #****************************************************************
   #
   #****************************************************************
    def check(self, classes):
        out = QStringList()
        for c in classes:
            if c: c.check(out)

        if out.count() > 0:
            msg = self.tr("Conversion is not possible.\nThe following programs are not found:\n - %1").arg(out.join('\n - '))
            raise Exception(unicode(msg))


   #****************************************************************
   # Error handler, terminate all worked threads
   #****************************************************************
    def threadError(self, trackIndex, msg):
        if trackIndex > -1:
            self._disc.track(trackIndex).setProgress(Statuses.Error, -1)

        self.stop()
        self.emit(SIGNAL("error(QString)"), msg)


   #****************************************************************
   #
   #****************************************************************
    def stop(self):
        if not self.isRunning():
            return

        for thread in self._threads:
            thread.stop()
            thread.wait()

        for track in self._disc.tracks():
            index = track.index()
            status = track.status()
            if ( status != Statuses.NotRunning and
                 status != Statuses.Error and
                 status != Statuses.OK
               ):
                track.setProgress(Statuses.Aborted, -1)

        self.emit(SIGNAL("finished()"))


   #****************************************************************
   #
   #****************************************************************
    def startThread(self):
        count = self._threadCount

        for thread in self._threads:
            if count == 0:
                break

            if thread.isFinished():
                continue

            if thread.isRunning():
                count -= 1
            else:
                if thread.isReadyStart():
                    thread.start()
                    count -= 1


   #****************************************************************
   #
   #****************************************************************
    def threadFinished(self):
        self.startThread()

        if not self.isRunning():
            del self._threads[:]
            self.emit(SIGNAL("finished()"))
            length = time.time() - self._startedTime
            print "Encoding time %d sec." %  length


   #****************************************************************
   #
   #****************************************************************
    def inputDataReady(self, trackIndex, fileName):
        self._disc.progressChanged(trackIndex, Statuses.OK, -1)


   #****************************************************************
   #
   #****************************************************************
    def isFinished(self):
        for thread in self._threads:
            if not thread.isFinished():
                return False

        return True


   #****************************************************************
   #
   #****************************************************************
    def isRunning(self):
        for thread in self._threads:
            if thread.isRunning():
                return True

        return False



class ConverterThread(QThread):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self):
        QThread.__init__(self)
        self._cancel = False
        self._trackIndex = -1


   #****************************************************************
   #
   #****************************************************************
    def trackIndex(self):
        return self._trackIndex


   #****************************************************************
   #
   #****************************************************************
    def stop(self):
        self._cancel = True


   #****************************************************************
   #
   #****************************************************************
    def inputDataReady(self, trackIndex, fileName):
        raise TypeError('Abstract method "ConverterThread.inputDataReady" called')


   #****************************************************************
   #
   #****************************************************************
    def run(self):
        try:
            self._run()

        except Abort, e:
            pass

        except Exception, e:
            self.emit(SIGNAL("error(int, QString)"), self._trackIndex, QString(unicode(e)))
            raise



class Splitter(ConverterThread):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self, trackList):
        ConverterThread.__init__(self)
        self._disc = trackList
        self._errors = QStringList()
        self._workDir = None


   #****************************************************************
   #
   #****************************************************************
    def isReadyStart(self):
        return True


   #****************************************************************
   # Split ape to temporary dir
   #****************************************************************
    def _run(self):
        prefTmpDir = qApp.preferences.value('Encoder/TmpDir').toString()

        if prefTmpDir.isEmpty():
            self._workDir = QFileInfo(self._disc.track(0).filePath()).dir()
        else:
            self._workDir = QDir(prefTmpDir + ('/flacon.%s' % os.getpid()))


        self.createDir(self._workDir)

        if not QFileInfo(self._workDir.path()).isWritable():
            e = self.tr("Can't write to directory \"%1\".").arg(QDir.cleanPath(self._workDir.path()))
            raise Exception(unicode(e))

        # shntool split -w -O always -n "%04d" -t "tmp%n" -d "~/music" 
        args = QStringList()
        args << qApp.preferences.value("Programs/shntool").toString()
        args << 'split'
        args << '-w'
        args << '-O' << 'always'
        args << '-n' << '%04d'
        args << '-t' << 'tmp%n'
        args << '-d' << self._workDir.path()
        args << self._disc.audioFileName()

        proc = tryPopen(args , stdin=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)

        self.sendCueData(proc)
        proc.stdin.close()
        self.parseOut(proc)
        res = proc.wait()
        if res:
            raise Exception(pipeToStr(proc.stderr))


   #****************************************************************
   # Create directory for temporary files
   #****************************************************************
    def createDir(self, workDir):
        if not workDir.mkpath('.'):
            e = self.tr("Can't create directory \"%1\".").arg(QDir.cleanPath(workDir.path()))
            raise Exception(unicode(e))


   #****************************************************************
   # Fix shnsplit error: m:ss.ff format can only be used with CD-quality files
   # I don't know last part in frames or in milliseconds?
   #****************************************************************
    def sendCueData(self, proc):
        cdQuality = (self._disc.cdQuality())

        cueData = self._disc.cueData()
        for i in xrange(0, cueData.count()):
            line = cueData[i].trimmed()
            if (not cdQuality) and (line.section(' ', 0, 0).toUpper() == 'INDEX'):
                ff = int(line.section(':', 2, 2))
                line = QString("%1.%2").arg(line.section(':', 0, 1)).arg(ff*10, 3, 10, QChar('0'))

            proc.stdin.write(line.toUtf8())
            proc.stdin.write("\n")


   #****************************************************************
   #
   #****************************************************************
    def parseOut(self, proc):
        codec = QTextCodec.codecForLocale()
        buff = ''
        trackRe = re.compile(r'/tmp(\d{1,4})\.wav\]\s+\(\d{1,2}:\d{2,2}\.\d{1,3}\)\s+:\s*\d+%')

        while True:
            if self._cancel:
                proc.terminate()
                proc.wait()
                raise Abort()

            c = proc.stderr.read(1)
            if not c: break
            buff += c
            #sys.stdout.write(c)

            #***************************************************
            if (c == "\n"):
                # Errorrs
                if (buff[8:16] == ': error:'):
                    self._errors << codec.toUnicode(buff[17:])

                buff = ''
            #***************************************************
            elif (c == '%'):
                percent = int((buff[-4:-1]))
                r = trackRe.search(buff)
                if (r):
                    self._trackIndex = int(r.groups()[0])-1

                if percent == 100:
                    fileName = QString('%1/tmp%2.wav').arg(self._workDir.path()).arg(self._trackIndex+1, 4, 10, QChar('0'))
                    if self._trackIndex > -1:
                        self.emit(SIGNAL("trackReady(int, QString)"), self._trackIndex, fileName)
                    else:
                        self.deleteFile(fileName)
                else:
                    self.emit(SIGNAL("progress(int, int, int)"), self._trackIndex, Statuses.Extracting, percent)


   #****************************************************************
   #
   #****************************************************************
    def deleteFile(self, fileName):
        f = QFile(fileName)
        if f.exists() and not f.remove():
            raise Exception(self.tr("Can't delete file:\n%1\n%2").arg(f.fileName()).arg(f.errorString()))


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        if not qApp.preferences.checkProgram('Programs/shntool'):
            out << 'shntool'



class Encoder(ConverterThread):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self):
        ConverterThread.__init__(self)
        self._wavFile = QString()
        self._outFile = QString()
        self._track = None
        self._proc = None


   #****************************************************************
   #
   #****************************************************************
    def track(self):
        return self._track


   #****************************************************************
   #
   #****************************************************************
    def setTrack(self, track):
        self._track = track
        self._trackIndex = track.index()
        self._outFile = track.filePath()


   #****************************************************************
   #
   #****************************************************************
    def inputDataReady(self, trackIndex, fileName):
        if self._trackIndex != trackIndex:
            return

        self.setWavFile(fileName)

        if self.isReadyStart():
            self.emit(SIGNAL("progress(int, int, int)"), self._trackIndex, Statuses.Queued, -1)
            self.emit(SIGNAL("readyStart()"))


   #****************************************************************
   #
   #****************************************************************
    def wavFile(self):
        return self._wavFile


   #****************************************************************
   #
   #****************************************************************
    def setWavFile(self, wavFile):
        self._wavFile = QString(wavFile)


   #****************************************************************
   #
   #****************************************************************
    def isReadyStart(self):
        return not self._wavFile.isEmpty() and self._track


   #****************************************************************
   #
   #****************************************************************
    def _run(self):
        self.emit(SIGNAL("progress(int, int, int)"), self._trackIndex, Statuses.Encoding, 0)

        if self._cancel: raise Abort()
        self.createDir()

        if self._cancel: raise Abort()
        self.encode()

        if self._cancel: raise Abort()
        self.deleteWavFile()

        self.emit(SIGNAL("trackReady(int, QString)"), self._trackIndex, self._outFile)


   #****************************************************************
   #
   #****************************************************************
    def createDir(self):
        outFile = QFile(self._track.filePath())
        outDir = QFileInfo(outFile).dir()

        if not outDir.mkpath('.'):
            raise Exception(self.tr("Can't create directory \"%1\".").arg(outDir.path()))

        if outFile.exists() and not outFile.remove():
            raise Exception(self.tr("Can't delete file:\n%1\n%2").arg(outFile.fileName()).arg(outFile.errorString()))


   #****************************************************************
   #
   #****************************************************************
    def encode(self):
        args = self.getEncoderArgs()
        #print unicode(args.join(" "))

        self._proc = tryPopen(args, stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=True)

        progress = 0

        total = os.path.getsize(unicode(self._wavFile))

        ready = 0
        buff = ''
        buffSize = int(total / 200)

        f = open(unicode(self._wavFile), 'rb')
        while True:

            if self._cancel:
                self._proc.terminate()
                self._proc.wait()
                raise Abort()

            buff = f.read(buffSize)
            if not buff: break

            ready+=buffSize
            p = int(ready *100 / total)
            if (progress != p):
                progress = p
                self.emit(SIGNAL("progress(int, int, int)"), self._trackIndex, Statuses.Encoding, progress)
            try:
                self._proc.stdin.write(buff)
            except:
                raise Exception(pipeToStr(self._proc.stderr))

        self._proc.stdin.close()
        res = self._proc.wait()
        if res:
            raise Exception(pipeToStr(self._proc.stderr))


   #****************************************************************
   #
   #****************************************************************
    def getEncoderArgs(self):
        raise TypeError('Abstract method "Encoder.getEncoderArgs" called')


   #****************************************************************
   #
   #****************************************************************
    def deleteWavFile(self):
        f = QFile(self._wavFile)
        if f.exists() and not f.remove():
            raise Exception(self.tr("Can't delete file:\n%1\n%2").arg(f.fileName()).arg(f.errorString()))



class WavEncoder(Encoder):
   #****************************************************************
   #
   #****************************************************************
    def encode(self):
        inFile = QFile(self._wavFile)
        if not inFile.rename(self._outFile):
            raise Exception(self.tr("Can't rename file:\n%1\n%2").arg(self._outFile).arg(inFile.errorString()))


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        pass


class FlacEncoder(Encoder):
   #****************************************************************
   #
   #****************************************************************
    def getEncoderArgs(self):
        args = QStringList()
        args << qApp.preferences.value("Programs/flac").toString()
        args << "--force"       # Force overwriting of output files.

        # Settings .................................................
        args << QString("--compression-level-%1").arg(qApp.preferences.value("Flac/Compression").toString())

        # Tags .....................................................
        if (self._track.artist()): args << "--tag" << QString("artist=%1").arg(self._track.artist())
        if (self._track.album()):  args << "--tag" << QString("album=%1").arg(self._track.album())
        if (self._track.genre()):  args << "--tag" << QString("genre=%1").arg(self._track.genre())
        if (self._track.year()):   args << "--tag" << QString("date=%1").arg(self._track.year())
        if (self._track.title()):  args << "--tag" << QString("title=%1").arg(self._track.title())
        if (self._track.discId()): args << "--tag" << QString("discId=%1").arg(self._track.discId())
        args << "--tag" << QString("tracknumber=%1").arg(self._track.trackNum())

        args << "-"
        args << "-o" << self._outFile,

        return args


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        if not qApp.preferences.checkProgram('Programs/flac'):
            out << 'flac'

class WVEncoder(Encoder):
   #****************************************************************
   #
   #****************************************************************
    def getEncoderArgs(self):
        args = QStringList()
        args << qApp.preferences.value("Programs/wavpack").toString()
        #args << "-y"       # "Yes" to all.

        # Settings .................................................
        compression = qApp.preferences.value("WavPack/Compression")
        if compression == 0:    args << "-f"
        if compression == 1:    args << "-h"
        if compression == 2:    args << "-hh"
                
        # Tags .....................................................
        if (self._track.artist()): args << "-w" << QString("Artist=%1").arg(self._track.artist())
        if (self._track.album()):  args << "-w" << QString("Album=%1").arg(self._track.album())
        if (self._track.genre()):  args << "-w" << QString("Genre=%1").arg(self._track.genre())
        if (self._track.year()):   args << "-w" << QString("Year=%1").arg(self._track.year())
        if (self._track.title()):  args << "-w" << QString("Title=%1").arg(self._track.title())
        #if (self._track.discId()): args << "-w" << QString("Disk=%1").arg(self._track.discId())
        args << "-w" << QString("Track=%1").arg(self._track.trackNum())

        args << "-"
        args << "-o" << self._outFile,

        
#        if DEBUG: print unicode(args.join(" "))
        return args


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        if not qApp.preferences.checkProgram('Programs/wavpack'):
            out << 'wavpack'



class OggEncoder(Encoder):
   #****************************************************************
   #
   #****************************************************************
    def getEncoderArgs(self):
        args = QStringList()
        args << qApp.preferences.value("Programs/oggenc").toString()
        args << "--quiet"

        # Quality settings .........................................
        if qApp.preferences.valueAsBool("Ogg/UseQuality"):
            args << "-q" << qApp.preferences.value("Ogg/Quality").toString()
        else:
            val = qApp.preferences.value("Ogg/NormBitrate").toString()
            if val:
                args << "-b" << val

            val = qApp.preferences.value("Ogg/MinBitrate").toString()
            if val:
                args << "-m" << val

            val = qApp.preferences.value("Ogg/MaxBitrate").toString()
            if val:
                args << "-M" << val

        # Tags .....................................................
        if (self._track.artist()): args << "--artist"  << self._track.artist()
        if (self._track.album()):  args << "--album"   << self._track.album()
        if (self._track.genre()):  args << "--genre"   << self._track.genre()
        if (self._track.year()):   args << "--date"    << str(self._track.year())
        if (self._track.title()):  args << "--title"   << self._track.title()
        if (self._track.discId()): args << "--comment" << QString("DISCID=%1").arg(self._track.discId())
        args << "--tracknum" << str(self._track.trackNum())

        # Files ....................................................
        args << "-o" << self._outFile,
        args << "-"
        return args


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        if not qApp.preferences.checkProgram('Programs/oggenc'):
            out << 'oggenc'


class Mp3Encoder(Encoder):
   #****************************************************************
   #
   #****************************************************************
    def getEncoderArgs(self):
        args = QStringList()
        args << qApp.preferences.value("Programs/lame").toString()
        args << "--silent"

        # Settings .................................................
        preset = qApp.preferences.value("Mp3/Preset").toString()

        if preset == 'avrMedium':
            args << "--preset" << "medium"

        elif preset == 'avrStandard':
            args << "--preset" << "standard"

        elif preset == 'avrStandardFast':
            args << "--preset" << "fast" << "standard"

        elif preset == 'avrExtreme':
            args << "--preset" << "extreme"

        elif preset == 'avrExtremeFast':
            args << "--preset" << "fast" << "extreme"

        elif preset == 'cbrInsane':
            args << "--preset" << "insane"

        elif preset == 'cbrKbps':
            args << "--preset" << "cbr" << qApp.preferences.value("Mp3/Bitrate").toString()

        elif preset == 'abrKbps':
            args << "--preset" << qApp.preferences.value("Mp3/Bitrate").toString()

        elif preset == 'vbrQuality':
            (quality, ok) = qApp.preferences.value("Mp3/Quality").toInt()
            args << "-V" << QString("%1").arg(9 - quality)


        # Replay Gain ...............................................
        if qApp.preferences.value("Mp3/ReplayGain") != ReplayGainTypes.Track:
            args << "--noreplaygain"

        # Tags .....................................................
        args << "--id3v2-only"
        if (self._track.artist()): args << "--ta" << self._track.artist()
        if (self._track.album()):  args << "--tl" << self._track.album()
        if (self._track.genre()):  args << "--tg" << self._track.genre()
        if (self._track.year()):   args << "--ty" << str(self._track.year())
        if (self._track.title()):  args << "--tt" << self._track.title()
        if (self._track.discId()): args << "--tc" << QString("discid=%1").arg(self._track.discId())
        args << "--tn" << str(self._track.trackNum())


        # Files ....................................................
        args << "-"
        args << self._outFile,

        return args


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        if not qApp.preferences.checkProgram('Programs/lame'):
            out << 'lame'



class Gain(ConverterThread):
   #****************************************************************
   #
   #****************************************************************
    def __init__(self, trackIndexses):
        ConverterThread.__init__(self)
        self._files = QStringList()
        self._trackIndexses = trackIndexses[:]


   #****************************************************************
   #
   #****************************************************************
    def inputDataReady(self, trackIndex, fileName):
        self._files << QStringList(fileName)
        self.emit(SIGNAL("progress(int, int, int)"), trackIndex, Statuses.WaitGain, -1)

        if self.isReadyStart():
            self.emit(SIGNAL("readyStart()"))


   #****************************************************************
   #
   #****************************************************************
    def isReadyStart(self):
        return self._files.count() == len(self._trackIndexses)# self._disc.tracksCount()


   #****************************************************************
   #
   #****************************************************************
    def _run(self):
        self._cancel = False
        args = self.getArgs()
        #print args.join(" ")

        proc = tryPopen(args, stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=True)

        self.parseOut(proc)
        res = proc.wait()
        if res:
            raise Exception(pipeToStr(proc.stderr))

        for i in xrange(0, len(self._trackIndexses)):
            self.emit(SIGNAL("trackReady(int, QString)"), self._trackIndexses[i], self._files[i])


   #****************************************************************
   #
   #****************************************************************
    def getArgs(self):
         raise TypeError('Abstract method "ReplayGain.getArgs" called')


   #****************************************************************
   #
   #****************************************************************
    def parseOut(self, proc):
         raise TypeError('Abstract method "ReplayGain.parseOut" called')



class FlacGain(Gain):
   #****************************************************************
   #
   #****************************************************************
    def getArgs(self):
        for i in self._trackIndexses:
            self.emit(SIGNAL("progress(int, int, int)"), i, Statuses.CalcGain, -1)

        args = QStringList()
        args << qApp.preferences.value("Programs/metaflac").toString()
        args << "--add-replay-gain"

        args << self._files

        return args


   #****************************************************************
   #
   #****************************************************************
    def parseOut(self, proc):
        pass


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        if not qApp.preferences.checkProgram('Programs/metaflac'):
            out << 'metaflac'



class WVGain(Gain):
   #****************************************************************
   #
   #****************************************************************
    def getArgs(self):
        for i in self._trackIndexses:
            self.emit(SIGNAL("progress(int, int, int)"), i, Statuses.CalcGain, -1)

        args = QStringList()
        args << qApp.preferences.value("Programs/wvgain").toString()
        args << "-a"

        args << self._files

#        if DEBUG: print unicode(args.join())
        return args


   #****************************************************************
   #
   #****************************************************************
    def parseOut(self, proc):
        pass


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        if not qApp.preferences.checkProgram('Programs/wvgain'):
            out << 'wvgain'



class OggGain(Gain):
   #****************************************************************
   #
   #****************************************************************
    def getArgs(self):
        args = QStringList()
        args << qApp.preferences.value("Programs/vorbisgain").toString()
        if qApp.preferences.value("Ogg/ReplayGain") == ReplayGainTypes.Album:
            args << "--album"

        args << self._files

        return args


   #****************************************************************
   #
   #****************************************************************
    def parseOut(self, proc):
        buff = ''
        trackName = ''

        calcRe = re.compile(r"^\s*(\d{1,3})%\s*-\s*(.*)")

        while True:
            if self._cancel:
                proc.terminate()
                proc.wait()
                raise Abort()

            c = proc.stderr.read(1)
            if not c: break
            buff += c
            #sys.stdout.write(c)

            if (c == "\r"):
                r = calcRe.search(buff)
                if (r):
                    if trackName != r.groups()[1]:
                        trackName = r.groups()[1]
                        n = self._files.indexOf(QString(trackName).trimmed())
                        if n>-1:
                            self._trackIndex = self._trackIndexses[n]

                    percent = int(r.groups()[0])
                    self.emit(SIGNAL("progress(int, int, int)"), self._trackIndex, Statuses.CalcGain, percent)

                buff = ''


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        if not qApp.preferences.checkProgram('Programs/vorbisgain'):
            out << 'vorbisgain'



class Mp3Gain(Gain):
   #****************************************************************
   #
   #****************************************************************
    def getArgs(self):
        args = QStringList()
        args << qApp.preferences.value("Programs/mp3gain").toString()
        args << "-a" # Album gain
        args << "-c" # ignore clipping warning when applying gain

        args << self._files

        return args


   #****************************************************************
   #
   #****************************************************************
    def parseOut(self, proc):
        buff = ''
        #trackName = ''
        #trackIndex = -1

        calcRe = re.compile(r"^\[(\d{1,4})/\d*]\s(\d{1,4})%")

        while True:
            if self._cancel:
                proc.terminate()
                proc.wait()
                raise Abort()

            c = proc.stderr.read(1)
            if not c: break
            buff += c
            #sys.stdout.write(c)

            if (c == "\r"):
                r = calcRe.search(buff)
                if (r):
                    n = self._trackIndexses[int(r.groups()[0])-1]
                    if self._trackIndex != n:
                        self.emit(SIGNAL("progress(int, int, int)"), self._trackIndex, Statuses.CalcGain, 100)
                        self._trackIndex = n

                    percent = int(r.groups()[1])
                    self.emit(SIGNAL("progress(int, int, int)"), self._trackIndex, Statuses.CalcGain, percent)

                buff = ''


   #****************************************************************
   #
   #****************************************************************
    @staticmethod
    def check(out):
        if not qApp.preferences.checkProgram('Programs/mp3gain'):
            out << 'mp3gain'




#*******************************************************************
#
#*******************************************************************
def getEncoderClass(audioFormat):
    if audioFormat == AudioFormat.FLAC: return FlacEncoder
    if audioFormat == AudioFormat.OGG:  return OggEncoder
    if audioFormat == AudioFormat.WAV:  return WavEncoder
    if audioFormat == AudioFormat.MP3:  return Mp3Encoder
    if audioFormat == AudioFormat.WV:   return WVEncoder

    common = QObject()
    raise Exception(common.tr("Unknown audio format."))



#*******************************************************************
#
#*******************************************************************
def getAlbumGainClass(audioFormat):

    if audioFormat == AudioFormat.FLAC:
        if qApp.preferences.value("Flac/ReplayGain") == ReplayGainTypes.Album:
            return FlacGain

    if audioFormat == AudioFormat.WV:
        if qApp.preferences.value("WavPack/ReplayGain") == ReplayGainTypes.Album:
            return WVGain

    if audioFormat == AudioFormat.OGG:
        if qApp.preferences.value("Ogg/ReplayGain") == ReplayGainTypes.Album:
            return OggGain

    if audioFormat == AudioFormat.MP3:
        if qApp.preferences.value("Mp3/ReplayGain") == ReplayGainTypes.Album:
            return Mp3Gain

    return


#*******************************************************************
#
#*******************************************************************
def getTrackGainClass(audioFormat):

    if audioFormat == AudioFormat.FLAC:
        if qApp.preferences.value("Flac/ReplayGain") == ReplayGainTypes.Track:
            return FlacGain

    if audioFormat == AudioFormat.WV:
        if qApp.preferences.value("WavPack/ReplayGain") == ReplayGainTypes.Track:
            return WVGain

    if audioFormat == AudioFormat.OGG:
        if qApp.preferences.value("Ogg/ReplayGain") == ReplayGainTypes.Track:
            return OggGain

    return
